#pragma once

#include <csrstr.h>
#include <csrmath.h>
#include <csrmthread.h>
#include <csrdomstore.h>
#include <csrfile.h>

class HInstance;
class HOpticalChain;

class HLightMap {
public:
	uint32_t width, height;
	double minX, minY, maxX, maxY;
	double* data;

public:
	HLightMap(const HLightMap& map);
	HLightMap(uint32_t width, uint32_t height);
	HLightMap(uint32_t width, uint32_t height, double minX, double minY, double maxX, double maxY);
	~HLightMap();

	void copy(const HLightMap& map);
	void copy2(const HLightMap& map);
	void copypart(const HLightMap& map);
	void setsize(uint32_t width, uint32_t height);

	inline double getX(int x) const {
		return minX + (maxX-minX)*x/width;
	}

	inline double getY(int y) const {
		return minY + (maxY-minY)*y/height;
	}

	void swapquarters();
};

class HOpticalElement {
	HOpticalChain* chain;
	csrutils::StringB nType;
	csrutils::StringB nTag;

public:
	HOpticalElement(const csrutils::String& nmType, const csrutils::String& nmTag);
	~HOpticalElement();

	inline csrutils::String& getTypeName() { return nType; }
	inline csrutils::String& getTagName() { return nTag; }
	inline void setTagName(const csrutils::String& nmTag) {
		nTag = nmTag;
	}

	virtual void run(HLightMap& input, HLightMap& output, csrutils::ThreadMonitor &mon) = 0;
	virtual void putInChain(HOpticalChain* chain) { this->chain = chain; }
	virtual HWND configWnd(HInstance* inst, HWND hParentWnd) { return NULL; }
	virtual bool getconfig(csrutils::DOMSObject& st) = 0;
	virtual bool setconfig(csrutils::DOMSObject& st) = 0;
};

struct HOpticalElementClass {
	const wchar_t* nmType;
	HOpticalElement* (*createElement) (const csrutils::String& name);
};

//class HOpticalElement_Config {
//}

struct HOpticalChain_Ctx;
class HOpticalChain {
	HOpticalChain_Ctx* ctx;
	HInstance* inst;

public:
	HOpticalChain(HInstance* inst);
	~HOpticalChain();

	void addElement(HOpticalElement* element, int pos);
	// Excludes element from chain and deletes it
	void removeElement(int pos);
	// Excludes element from chain without deleting it
	HOpticalElement* excludeElement(int pos);
	HOpticalElement* getElement(int pos);
	HOpticalElement* getElementByName(const csrutils::String& name);
	int countElements();
	int positionElement(HOpticalElement* element);
	void clear();

	void render(csrutils::ThreadMonitor& mon); //(HLightMap& img, double Z);
	void render_proc(csrutils::ThreadMonitor& mon);

	bool getconfig(csrutils::DOMSObject& st);
	bool setconfig(csrutils::DOMSObject& st);

	bool open(csrutils::File& file);
	bool save(csrutils::File& file);
};

class HVideo {
public:
	virtual bool getframe(HLightMap* map) = 0;
	virtual uint32_t getframescount() = 0;
	virtual uint32_t getframeid() = 0;
	virtual bool nextframe() = 0;
	virtual bool seekframe(uint32_t frameid) = 0;
	virtual HLightMap* getmap() = 0;
};

struct HChainVideo_Ctx;
class HChainVideo : public HVideo {
	HChainVideo_Ctx* ctx;
	HOpticalChain* chain;

public:
	HChainVideo(HOpticalChain* chain);
	~HChainVideo();

	bool getframe(HLightMap* map);
	uint32_t getframescount();
	uint32_t getframeid();
	bool nextframe();
	bool seekframe(uint32_t frameid);
	HLightMap* getmap();
};

struct HAnimator_Ctx;
class HAnimator {
	HAnimator_Ctx* ctx;
	HInstance* inst;
	HOpticalChain* chain;

public:
	HAnimator(HInstance* inst, HOpticalChain* chain);
	~HAnimator();

	void run(HVideo* video);
	void render_lightvideo(HVideo* video);
};

struct HInstance_Ctx;
class HInstance {
	HInstance_Ctx* ctx;
	HINSTANCE hInst;

public:
	HInstance(HINSTANCE hInst);
	~HInstance();

	inline HINSTANCE getHInst() { return hInst; };
	void regElementClass(const HOpticalElementClass* cls);
	int getElementClasses(const HOpticalElementClass** clses);
	const HOpticalElementClass* getElementClass(const wchar_t* name);
};

struct HChainView_Ctx;
class HChainView {
	HChainView_Ctx* ctx;
	HInstance* inst;
	HOpticalChain* chain;

public:
	HChainView(HInstance* inst, HOpticalChain* chain);
	~HChainView();

	void render();

#ifdef _HUYGENS_INT_
	void _paint(HDC hdc);
	LRESULT _wndproc(HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam);
	LRESULT _wndproc_notify(HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam);
	LRESULT _insertdlgproc(HWND hDlg, UINT message, WPARAM wParam, LPARAM lParam);
	LRESULT _elconfdlgproc(HWND hDlg, UINT message, WPARAM wParam, LPARAM lParam);
	LRESULT _elmovedlgproc(HWND hDlg, UINT message, WPARAM wParam, LPARAM lParam);
	LRESULT _settingsdlgproc(HWND hDlg, UINT message, WPARAM wParam, LPARAM lParam);
	bool _initListView();
	bool _dispinfoListView(LPARAM lParam);
	void _chainsave();
	void _chainopen(bool def=false);
#endif
};

struct HImageView_Ctx;
class HImageView {
#ifdef _HUYGENS_INT_
	public:
#endif
	HImageView_Ctx* ctx;
#ifdef _HUYGENS_INT_
	private:
#endif
	HInstance* inst;
	HLightMap* map;
	double amplif;

public:
	HImageView(HInstance* inst, HLightMap* map, HWND hWndParent, bool asChild=false);
	HImageView(HInstance* inst, HWND hWndParent, bool asChild=false);
	~HImageView();

	void update();
	void openfile(csrutils::File file);

#ifdef _HUYGENS_INT_
	void _paint(HDC hdc);
	LRESULT _wndproc(HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam);
	void _initcontrols();
	void _hidecontrols();
	HWND getwnd();
#endif
};

struct HVideoView_Ctx;
class HVideoView {
	HVideoView_Ctx* ctx;
	HVideo* video;
	HInstance* inst;

public:
	HVideoView(HInstance* inst, HVideo* video, HWND hWndParent, bool asChild=false);
	~HVideoView();

#ifdef _HUYGENS_INT_
	void _paint(HDC hdc);
	LRESULT _wndproc(HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam);
	void _vidseek(float p);
	//void _initcontrols();
	//void _hidecontrols();
#endif
};

enum HRenderType {
	HRT_AMPL = 1,
	HRT_COLPHASE
};

void Huygens_RenderLightMap(HLightMap* map, csrutils::Image* img, double amplif, HRenderType type);
void Huygens_CopyWithScale(HDC dst, HDC src, uint32_t dst_w, uint32_t dst_h, uint32_t src_w,  uint32_t src_h);

#ifdef _HUYGENS_INT_
#include "huygens_int.h"
#endif
