#define _HUYGENS_INT_
#define _CRT_SECURE_NO_WARNINGS
#include "stdafx.h"
#include "hoptelems.h"
#include <map>
#include <csrmath.h>
#include <csrmath_complex.h>
#include "resource.h"

using namespace csrutils;
using csrmath::ComplexD;
using csrmath::expi;


void registerStdOpticalElements(HInstance* inst) {
	inst->regElementClass( getHWaveViewClass() );
	inst->regElementClass( getHPlanarWaveClass() );
	inst->regElementClass( getHBesselWaveClass() );
	inst->regElementClass( getHDifrDotClass() );
	inst->regElementClass( getHZonePlateClass() );
	inst->regElementClass( getHBesselPlateClass() );
	inst->regElementClass( getHThinLensClass() );
	inst->regElementClass( getHImageTranspClass() );
	inst->regElementClass( getHSpaceHuygensClass() );
	inst->regElementClass( getHSpaceFresnelClass() );
	inst->regElementClass( getHSpaceFourierClass() );
	inst->regElementClass( getHSpaceShiftClass() );
}


HOpticalElement::HOpticalElement(const String& nmType, const String& nmTag) {
	nType = nmType;
	nTag  = nmTag;
}

HOpticalElement::~HOpticalElement() {
}


/**************************************
 **         Class HWaveView          **
 **************************************/

HOpticalElementClass *clsHWaveView = 0;
const wchar_t* szClsHWaveView = L"WaveView";

HOpticalElement* createHWaveView(const String& name) {
	return new HWaveView(name);
}

const HOpticalElementClass* getHWaveViewClass() {
	if( clsHWaveView==0 ) {
		clsHWaveView = new HOpticalElementClass;
		clsHWaveView->nmType = szClsHWaveView;
		clsHWaveView->createElement = createHWaveView;
	}
	return clsHWaveView;
}

HWaveView::HWaveView(const String& nmTag) : HOpticalElement(szClsHWaveView,nmTag) {
	view = 0;
	map = new HLightMap(512,512);
	iwidth  = 0;
	iheight = 0;
}

void HWaveView::run(HLightMap& input, HLightMap& output, csrutils::ThreadMonitor &mon) {
	output.copy(input);
	if( iwidth!=0 && iheight!=0 )
		map->copypart(input);
	else
		map->copy2(input);
	if( view!=0 )
		view->update();
}

HWND HWaveView::configWnd(HInstance *inst, HWND hParentWnd) {
	if( view!=0 )
		delete view;
	view = new HImageView(inst,map,hParentWnd);
	return 0;
}

bool HWaveView::getconfig(csrutils::DOMSObject& st) {
	st[L"iwidth"]  = iwidth;
	st[L"iheight"] = iheight;
	return true;
}

bool HWaveView::setconfig(csrutils::DOMSObject& st) {
	if( st[L"iwidth"].exsists() )
		iwidth = st[L"iwidth"];
	else
		iwidth = 0;
	if( st[L"iheight"].exsists() )
		iheight = st[L"iheight"];
	else
		iheight = 0;
	if( iwidth!=0 && iheight!=0 )
		map->setsize(iwidth,iheight);
	return true;
}

HLightMap* HWaveView::getmap() {
	return map;
}



/**************************************
 **          Class HDifrDot          **
 **************************************/

HOpticalElementClass *clsHDifrDot = 0;
const wchar_t* szClsHDifrDot = L"DifrDot";

HOpticalElement* createHDifrDot(const String& name) {
	return new HDifrDot(name);
}

const HOpticalElementClass* getHDifrDotClass() {
	if( clsHDifrDot==0 ) {
		clsHDifrDot = new HOpticalElementClass;
		clsHDifrDot->nmType = szClsHDifrDot;
		clsHDifrDot->createElement = createHDifrDot;
	}
	return clsHDifrDot;
}

HDifrDot::HDifrDot(const String& nmTag)
		: HOpticalElement(szClsHDifrDot,nmTag), radius(0.1) {
}

HDifrDot::HDifrDot(const String& nmTag, double radius)
		: HOpticalElement(szClsHDifrDot,nmTag), radius(radius) {
}

void HDifrDot::run(HLightMap& input, HLightMap& output, csrutils::ThreadMonitor &mon) {
	double r2 = radius*radius;
	for( uint32_t x=0; x<output.width; x++ )
	for( uint32_t y=0; y<output.width; y++ ) {
		double X = input.getX(x);
		double Y = input.getY(y);
		if( X*X + Y*Y < r2 ) {
			output.data[ 2*(x+y*output.width) + 0 ] = input.data[ 2*(x+y*output.width) + 0 ];
			output.data[ 2*(x+y*output.width) + 1 ] = input.data[ 2*(x+y*output.width) + 1 ];
		} else {
			output.data[ 2*(x+y*output.width) + 0 ] = 0.0;
			output.data[ 2*(x+y*output.width) + 1 ] = 0.0;
		}
	}
}

bool HDifrDot::getconfig(csrutils::DOMSObject& st) {
	st[L"radius"] = radius;
	return true;
}

bool HDifrDot::setconfig(csrutils::DOMSObject& st) {
	radius = st[L"radius"];
	return true;
}




/**************************************
 **        Class HZonePlate          **
 **************************************/

HOpticalElementClass *clsHZonePlate = 0;
const wchar_t* szClsHZonePlate = L"ZonePlate";

HOpticalElement* createHZonePlate(const String& name) {
	return new HZonePlate(name);
}

const HOpticalElementClass* getHZonePlateClass() {
	if( clsHZonePlate==0 ) {
		clsHZonePlate = new HOpticalElementClass;
		clsHZonePlate->nmType = szClsHZonePlate;
		clsHZonePlate->createElement = createHZonePlate;
	}
	return clsHZonePlate;
}

HZonePlate::HZonePlate(const String& nmTag)
		: HOpticalElement(szClsHZonePlate,nmTag), wavelength(0.0005), focuslength(10.0) {
}

HZonePlate::HZonePlate(const String& nmTag, double focuslength, double wavelength)
		: HOpticalElement(szClsHZonePlate,nmTag), focuslength(focuslength), wavelength(wavelength) {
}

void HZonePlate::run(HLightMap& input, HLightMap& output, csrutils::ThreadMonitor &mon) {
	for( uint32_t x=0; x<output.width; x++ )
	for( uint32_t y=0; y<output.width; y++ ) {
		double X = input.getX(x);
		double Y = input.getY(y);
		int m = (int) floor( ( X*X+Y*Y ) / (focuslength*wavelength) );
		if( m%2==0 ) {
			output.data[ 2*(x+y*output.width) + 0 ] = input.data[ 2*(x+y*output.width) + 0 ];
			output.data[ 2*(x+y*output.width) + 1 ] = input.data[ 2*(x+y*output.width) + 1 ];
		} else {
			output.data[ 2*(x+y*output.width) + 0 ] = 0.0;
			output.data[ 2*(x+y*output.width) + 1 ] = 0.0;
		}
	}
}

bool HZonePlate::getconfig(csrutils::DOMSObject& st) {
	st[L"focuslength"] = focuslength;
	st[L"wavelength"]  = wavelength;
	return true;
}

bool HZonePlate::setconfig(csrutils::DOMSObject& st) {
	focuslength = st[L"focuslength"];
	wavelength  = st[L"wavelength"];
	return true;
}





/**************************************
 **         Class HThinLens          **
 **************************************/

HOpticalElementClass *clsHThinLens = 0;
const wchar_t* szClsHThinLens = L"ThinLens";

HOpticalElement* createHThinLens(const String& name) {
	return new HThinLens(name);
}

const HOpticalElementClass* getHThinLensClass() {
	if( clsHThinLens==0 ) {
		clsHThinLens = new HOpticalElementClass;
		clsHThinLens->nmType = szClsHThinLens;
		clsHThinLens->createElement = createHThinLens;
	}
	return clsHThinLens;
}

HThinLens::HThinLens(const String& nmTag)
		: HOpticalElement(szClsHThinLens,nmTag), focuslength(10.0), wavelength(0.0005) {
}

HThinLens::HThinLens(const String& nmTag, double focuslength)
		: HOpticalElement(szClsHThinLens,nmTag), focuslength(focuslength), wavelength(wavelength) {
}

void HThinLens::run(HLightMap& input, HLightMap& output, csrutils::ThreadMonitor &mon) {
	double K = 2.0*M_PI / wavelength;
	for( uint32_t x=0; x<output.width; x++ )
	for( uint32_t y=0; y<output.width; y++ ) {
		double X = output.getX(x);
		double Y = output.getY(y);
		*(ComplexD*) ( &output.data[ 2*(x+y*output.width) ] ) =
			expi( - K / (2.0*focuslength) * ( X*X+Y*Y ) ) *
			*(ComplexD*) ( &input.data[ 2*(x+y*output.width) ] );
			//output.data[ 2*(x+y*output.width) + 0 ] = input.data[ 2*(x+y*output.width) + 0 ];
			//output.data[ 2*(x+y*output.width) + 1 ] = input.data[ 2*(x+y*output.width) + 1 ];
		
	}
}

bool HThinLens::getconfig(csrutils::DOMSObject& st) {
	st[L"focuslength"] = focuslength;
	st[L"wavelength"]  = wavelength;
	return true;
}

bool HThinLens::setconfig(csrutils::DOMSObject& st) {
	focuslength = st[L"focuslength"];
	wavelength  = st[L"wavelength"];
	return true;
}




/**************************************
 **        Class HImageTransp        **
 **************************************/

HOpticalElementClass *clsHImageTransp = 0;
const wchar_t* szClsHImageTransp = L"ImageTransp";

HOpticalElement* createHImageTransp(const String& name) {
	return new HImageTransp(name);
}

const HOpticalElementClass* getHImageTranspClass() {
	if( clsHImageTransp==0 ) {
		clsHImageTransp = new HOpticalElementClass;
		clsHImageTransp->nmType = szClsHImageTransp;
		clsHImageTransp->createElement = createHImageTransp;
	}
	return clsHImageTransp;
}

HImageTransp::HImageTransp(const String& nmTag) : HOpticalElement(szClsHImageTransp,nmTag) {
	//image = Image::load(L"C:\\Users\\Victor\\Documents\\NSU\\Education\\ 2\\SCREW_1b.bmp",ImageFileFormat_BMP);
	image = NULL;
	img_w = 0.5;
	img_h = 0.5;
}

HImageTransp::HImageTransp(const String& nmTag, csrutils::Image* _image)
		: HOpticalElement(szClsHImageTransp,nmTag), image(_image) {
}

void HImageTransp::run(HLightMap& input, HLightMap& output, csrutils::ThreadMonitor &mon) {
	ComplexD M_I(0.0,1.0);
	ComplexD* input_data  = (ComplexD*) input.data;
	ComplexD* output_data = (ComplexD*) output.data;
	Image* img;
	if( image==NULL ) {
		if( filepath==L"" ) {
			output.copy(input);
			return;
		}
		// img = Image::load(filepath.bytes(),ImageFileFormat_BMP);
		img = new Image;  
		img->load(filepath,ImageFileFormat_BMP);
	}
	else
		img = image;
	for( uint32_t x=0; x<output.width; x++ )
	for( uint32_t y=0; y<output.width; y++ ) {
		double X = input.getX(x);
		double Y = input.getY(y);
		int x0 = (int)( ( 0.5 + X / img_w ) * img->getWidth() );
		int y0 = (int)( ( 0.5 + Y / img_h ) * img->getHeight() );
		uint32_t c = img->getPixel(x0,y0);
		if( c==0 )
			output_data[x+y*output.width] = - input_data[x+y*output.width];
		else
			output_data[x+y*output.width] = input_data[x+y*output.width];
	}
	delete img;
}

bool HImageTransp::getconfig(csrutils::DOMSObject& st) {
	st[L"img_w"] = img_w;
	st[L"img_h"] = img_h;
	st[L"file"] = filepath;
	return true;
}

bool HImageTransp::setconfig(csrutils::DOMSObject& st) {
	img_w = st[L"img_w"];
	img_h = st[L"img_h"];
	filepath = st[L"file"];
	return true;
}
